==> isolate_contactor-2.1.0.tar.gz: lib/src/isolate_contactor/web_platform/isolate_contactor_web_worker.dart <==
import 'dart:async';
import 'dart:html';

import 'package:isolate_contactor/src/isolate_contactor.dart';
import 'package:isolate_contactor/src/isolate_contactor_controller/isolate_contactor_controller_web.dart';

import '../../isolate_contactor_controller.dart';
import '../../utils/exception.dart';
import '../../utils/utils.dart';
import '../isolate_contactor_web.dart';

class IsolateContactorInternalWorker<T> implements IsolateContactorInternal<T> {
  /// For debugging
  bool _debugMode = false;

  /// Check for current isolate in bool
  bool _isComputing = false;

  /// Check for current cumputing state in enum with listener
  final StreamController<ComputeState> _computeStateStreamController =
      StreamController.broadcast();

  /// Check for current cumputing state in enum with listener
  final StreamController<T> _mainStreamController =
      StreamController.broadcast();

  /// Listener for result
  IsolateContactorController<T>? _isolateContactorController;

  // final _isolateWorker = Worker("isolate.dart.js");

  /// Control the function of isolate
  // ignore: unused_field
  late void Function(dynamic) _isolateFunction;

  /// Control the parameters of isolate
  // ignore: unused_field
  late dynamic _isolateParam;

  late String _workerName;

  late T Function(dynamic) _converter;
  late T Function(dynamic) _workerConverter;

  /// Create an instance
  IsolateContactorInternalWorker._({
    required FutureOr<void> Function(dynamic) isolateFunction,
    required dynamic isolateParam,
    required String workerName,
    required T Function(dynamic) converter,
    required T Function(dynamic) workerConverter,
    bool debugMode = false,
  }) {
    _debugMode = debugMode;
    _isolateFunction = isolateFunction;
    _workerName = workerName;
    _converter = converter;
    _workerConverter = workerConverter;
    _isolateParam = isolateParam;
  }

  /// Create an instance
  static Future<IsolateContactorInternalWorker<T>> create<T>({
    required FutureOr<T> Function(dynamic) function,
    required String workerName,
    required T Function(dynamic) converter,
    required T Function(dynamic) workerConverter,
    bool debugMode = true,
  }) async {
    IsolateContactorInternalWorker<T> isolateContactor =
        IsolateContactorInternalWorker._(
      isolateFunction: internalIsolateFunction,
      workerName: workerName,
      isolateParam: function,
      converter: converter,
      workerConverter: workerConverter,
      debugMode: debugMode,
    );

    await isolateContactor._initial();

    return isolateContactor;
  }

  /// Create modified isolate function
  static Future<IsolateContactorInternalWorker<T>> createOwnIsolate<T>({
    required void Function(dynamic) isolateFunction,
    required String workerName,
    required dynamic initialParams,
    required T Function(dynamic) converter,
    required T Function(dynamic) workerConverter,
    bool debugMode = false,
  }) async {
    IsolateContactorInternalWorker<T> isolateContactor =
        IsolateContactorInternalWorker._(
      isolateFunction: isolateFunction,
      workerName: workerName,
      isolateParam: initialParams ?? [],
      converter: converter,
      workerConverter: workerConverter,
      debugMode: debugMode,
    );

    await isolateContactor._initial();

    return isolateContactor;
  }

  /// Initialize
  Future<void> _initial() async {
    _isolateContactorController = IsolateContactorControllerImpl(
      Worker("$_workerName.js"),
      converter: _converter,
      workerConverter: _workerConverter,
    );
    _isolateContactorController!.onMessage.listen((message) {
      _printDebug('[Main Stream] Message received from Worker: $message');
      _computeStateStreamController.sink.add(ComputeState.computed);
      _mainStreamController.sink.add(message);
      _isComputing = false;
    }).onError((err, stack) {
      _printDebug('[Main Stream] Error message received from Worker: $err');
      _computeStateStreamController.sink.add(ComputeState.computed);
      _mainStreamController.sink.addError(err, stack);
      _isComputing = false;
    });

    _isComputing = false;
    _computeStateStreamController.sink.add(ComputeState.computed);
    _printDebug('Initialized');
  }

  /// Get current message as stream
  @override
  Stream<T> get onMessage => _mainStreamController.stream;

  /// Get current state
  @override
  Stream<ComputeState> get onComputeState =>
      _computeStateStreamController.stream;

  /// Is current isolate computing
  @override
  bool get isComputing => _isComputing;

  /// Restart current [Isolate]
  ///
  /// Umplemented in web platform at the moment.
  @override
  Future<void> restart() async {
    if (_isolateContactorController == null) {
      _printDebug('! This isolate has been terminated');
      return;
    }

    await _isolateContactorController!.close();

    await _initial();
  }

  @override
  Future<void> close() => dispose();

  @override
  Future<void> terminate() => dispose();

  /// Dispose current [Isolate]
  @override
  Future<void> dispose() async {
    _isComputing = false;
    _isolateContactorController?.sendIsolate(IsolateState.dispose);
    _computeStateStreamController.sink.add(ComputeState.computed);

    _computeStateStreamController.close;
    await _isolateContactorController?.close();
    await _mainStreamController.close();

    _isolateContactorController = null;

    _printDebug('Disposed');
  }

  /// Send message to child isolate [function].
  ///
  /// Throw IsolateContactorException if error occurs.
  @override
  Future<T> sendMessage(dynamic message) {
    if (_isolateContactorController == null) {
      _printDebug('! This isolate has been terminated');
      return throw IsolateException(
        'This isolate was terminated',
        StackTrace.empty,
      );
    }

    if (_isComputing) {
      _printDebug(
          '! This isolate is still being computed, so the current request has been revoked!');

      return throw IsolateException(
        'This isolate is still being computed, so the current request has been revoked',
        StackTrace.empty,
      );
    }

    _isComputing = true;
    _computeStateStreamController.sink.add(ComputeState.computing);

    final Completer<T> completer = Completer();
    StreamSubscription? sub;
    sub = _isolateContactorController!.onMessage.listen((result) async {
      if (!completer.isCompleted) {
        completer.complete(result);
        await sub?.cancel();
      }
    })
      ..onError((err, stack) async {
        completer.completeError(err, stack);
        await sub?.cancel();
      });

    _printDebug('Message send to isolate: $message');

    _isolateContactorController!.sendIsolate(message);

    return completer.future;
  }

  /// Print if [debugMode] is true
  void _printDebug(Object? object, [bool force = false]) {
    // ignore: avoid_print
    if (_debugMode && !force) {
      print('[${IsolateContactor.debugLogPrefix}]: $object');
    }
  }
}

==> isolate_contactor-4.0.1.tar.gz: lib/src/isolate_contactor/web_platform/isolate_contactor_web_worker.dart <==
import 'dart:async';
import 'dart:html';

import 'package:isolate_contactor/src/isolate_contactor.dart';
import 'package:isolate_contactor/src/isolate_contactor_controller/isolate_contactor_controller_web.dart';

import '../../isolate_contactor_controller.dart';
import '../../utils/exception.dart';
import '../../utils/utils.dart';
import '../isolate_contactor_web.dart';

class IsolateContactorInternalWorker<R, P>
    implements IsolateContactorInternal<R, P> {
  /// For debugging
  bool _debugMode = false;

  /// Check for current isolate in bool
  bool _isComputing = false;

  /// Check for current cumputing state in enum with listener
  final StreamController<ComputeState> _computeStateStreamController =
      StreamController.broadcast();

  /// Check for current cumputing state in enum with listener
  final StreamController<R> _mainStreamController =
      StreamController.broadcast();

  /// Listener for result
  IsolateContactorController<R, P>? _isolateContactorController;

  // final _isolateWorker = Worker("isolate.dart.js");

  /// Control the function of isolate
  // ignore: unused_field
  late void Function(dynamic) _isolateFunction;

  /// Control the parameters of isolate
  // ignore: unused_field
  late dynamic _isolateParam;

  late String _workerName;

  late R Function(dynamic) _converter;
  late R Function(dynamic) _workerConverter;

  /// Create an instance
  IsolateContactorInternalWorker._({
    required FutureOr<void> Function(dynamic) isolateFunction,
    required Object? isolateParam,
    required String workerName,
    required R Function(dynamic) converter,
    required R Function(dynamic) workerConverter,
    bool debugMode = false,
  }) {
    _debugMode = debugMode;
    _isolateFunction = isolateFunction;
    _workerName = workerName;
    _converter = converter;
    _workerConverter = workerConverter;
    _isolateParam = isolateParam;
  }

  /// Create an instance
  static Future<IsolateContactorInternalWorker<R, P>> create<R, P>({
    required FutureOr<R> Function(P params) function,
    required String workerName,
    required R Function(dynamic) converter,
    required R Function(dynamic) workerConverter,
    bool debugMode = true,
  }) async {
    IsolateContactorInternalWorker<R, P> isolateContactor =
        IsolateContactorInternalWorker._(
      isolateFunction: internalIsolateFunction,
      workerName: workerName,
      isolateParam: function,
      converter: converter,
      workerConverter: workerConverter,
      debugMode: debugMode,
    );

    await isolateContactor._initial();

    return isolateContactor;
  }

  /// Create modified isolate function
  static Future<IsolateContactorInternalWorker<R, P>> createOwnIsolate<R, P>({
    required void Function(dynamic) isolateFunction,
    required String workerName,
    required Object? initialParams,
    required R Function(dynamic) converter,
    required R Function(dynamic) workerConverter,
    bool debugMode = false,
  }) async {
    IsolateContactorInternalWorker<R, P> isolateContactor =
        IsolateContactorInternalWorker._(
      isolateFunction: isolateFunction,
      workerName: workerName,
      isolateParam: initialParams,
      converter: converter,
      workerConverter: workerConverter,
      debugMode: debugMode,
    );

    await isolateContactor._initial();

    return isolateContactor;
  }

  /// Initialize
  Future<void> _initial() async {
    _isolateContactorController = IsolateContactorControllerImpl(
      Worker("$_workerName.js"),
      converter: _converter,
      workerConverter: _workerConverter,
    );
    _isolateContactorController!.onMessage.listen((message) {
      _printDebug('[Main Stream] Message received from Worker: $message');
      _computeStateStreamController.sink.add(ComputeState.computed);
      _mainStreamController.sink.add(message);
      _isComputing = false;
    }).onError((err, stack) {
      _printDebug('[Main Stream] Error message received from Worker: $err');
      _computeStateStreamController.sink.add(ComputeState.computed);
      _mainStreamController.sink.addError(err, stack);
      _isComputing = false;
    });

    _isComputing = false;
    _computeStateStreamController.sink.add(ComputeState.computed);
    _printDebug('Initialized');
  }

  /// Get current message as stream
  @override
  Stream<R> get onMessage => _mainStreamController.stream;

  /// Get current state
  @override
  Stream<ComputeState> get onComputeState =>
      _computeStateStreamController.stream;

  /// Is current isolate computing
  @override
  bool get isComputing => _isComputing;

  /// Restart current [Isolate]
  ///
  /// Umplemented in web platform at the moment.
  @override
  Future<void> restart() async {
    if (_isolateContactorController == null) {
      _printDebug('! This isolate has been terminated');
      return;
    }

    await _isolateContactorController!.close();

    await _initial();
  }

  @override
  Future<void> close() => dispose();

  @override
  Future<void> terminate() => dispose();

  /// Dispose current [Isolate]
  @override
  Future<void> dispose() async {
    _isComputing = false;
    _isolateContactorController?.sendIsolateState(IsolateState.dispose);
    _computeStateStreamController.sink.add(ComputeState.computed);

    _computeStateStreamController.close;
    await _isolateContactorController?.close();
    await _mainStreamController.close();

    _isolateContactorController = null;

    _printDebug('Disposed');
  }

  /// Send message to child isolate [function].
  ///
  /// Throw IsolateContactorException if error occurs.
  @override
  Future<R> sendMessage(P message) {
    if (_isolateContactorController == null) {
      _printDebug('! This isolate has been terminated');
      return throw IsolateException(
        'This isolate was terminated',
        StackTrace.empty,
      );
    }

    if (_isComputing) {
      _printDebug(
          '! This isolate is still being computed, so the current request has been revoked!');

      return throw IsolateException(
        'This isolate is still being computed, so the current request has been revoked',
        StackTrace.empty,
      );
    }

    _isComputing = true;
    _computeStateStreamController.sink.add(ComputeState.computing);

    final Completer<R> completer = Completer();
    StreamSubscription? sub;
    sub = _isolateContactorController!.onMessage.listen((result) async {
      if (!completer.isCompleted) {
        completer.complete(result);
        await sub?.cancel();
      }
    })
      ..onError((err, stack) async {
        completer.completeError(err, stack);
        await sub?.cancel();
      });

    _printDebug('Message send to isolate: $message');

    _isolateContactorController!.sendIsolate(message);

    return completer.future;
  }

  /// Print if [debugMode] is true
  void _printDebug(Object? object, [bool force = false]) {
    // ignore: avoid_print
    if (_debugMode && !force) {
      print('[${IsolateContactor.debugLogPrefix}]: $object');
    }
  }
}
